@@ -186,7 +186,7 @@ pub fn gen_py_method(
     check_generic(sig)?;
     ensure_not_async_fn(sig)?;
     ensure_function_options_valid(&options)?;
-    let method = PyMethod::parse(sig, &mut *meth_attrs, options)?;
+    let method = PyMethod::parse(sig, meth_attrs, options)?;
     let spec = &method.spec;
 
     Ok(match (method.kind, &spec.tp) {
@@ -196,7 +196,7 @@ pub fn gen_py_method(
             GeneratedPyMethod::Method(impl_py_class_attribute(cls, spec)?)
         }
         (PyMethodKind::Proto(proto_kind), _) => {
-            ensure_no_forbidden_protocol_attributes(spec, &method.method_name)?;
+            ensure_no_forbidden_protocol_attributes(&proto_kind, spec, &method.method_name)?;
             match proto_kind {
                 PyMethodProtoKind::Slot(slot_def) => {
                     let slot = slot_def.generate_type_slot(cls, spec, &method.method_name)?;
@@ -206,7 +206,7 @@ pub fn gen_py_method(
                     GeneratedPyMethod::Proto(impl_call_slot(cls, method.spec)?)
                 }
                 PyMethodProtoKind::Traverse => {
-                    GeneratedPyMethod::Proto(impl_traverse_slot(cls, method.spec))
+                    GeneratedPyMethod::Proto(impl_traverse_slot(cls, spec.name))
                 }
                 PyMethodProtoKind::SlotFragment(slot_fragment_def) => {
                     let proto = slot_fragment_def.generate_pyproto_fragment(cls, spec)?;
@@ -215,27 +215,44 @@ pub fn gen_py_method(
             }
         }
         // ordinary functions (with some specialties)
-        (_, FnType::Fn(_)) => GeneratedPyMethod::Method(impl_py_method_def(cls, spec, None)?),
+        (_, FnType::Fn(_)) => GeneratedPyMethod::Method(impl_py_method_def(
+            cls,
+            spec,
+            &spec.get_doc(meth_attrs),
+            None,
+        )?),
         (_, FnType::FnClass) => GeneratedPyMethod::Method(impl_py_method_def(
             cls,
             spec,
+            &spec.get_doc(meth_attrs),
             Some(quote!(_pyo3::ffi::METH_CLASS)),
         )?),
         (_, FnType::FnStatic) => GeneratedPyMethod::Method(impl_py_method_def(
             cls,
             spec,
+            &spec.get_doc(meth_attrs),
             Some(quote!(_pyo3::ffi::METH_STATIC)),
         )?),
         // special prototypes
-        (_, FnType::FnNew) => GeneratedPyMethod::Proto(impl_py_method_def_new(cls, spec)?),
+        (_, FnType::FnNew) | (_, FnType::FnNewClass) => {
+            GeneratedPyMethod::Proto(impl_py_method_def_new(cls, spec)?)
+        }
 
         (_, FnType::Getter(self_type)) => GeneratedPyMethod::Method(impl_py_getter_def(
             cls,
-            PropertyType::Function { self_type, spec },
+            PropertyType::Function {
+                self_type,
+                spec,
+                doc: spec.get_doc(meth_attrs),
+            },
         )?),
         (_, FnType::Setter(self_type)) => GeneratedPyMethod::Method(impl_py_setter_def(
             cls,
-            PropertyType::Function { self_type, spec },
+            PropertyType::Function {
+                self_type,
+                spec,
+                doc: spec.get_doc(meth_attrs),
+            },
         )?),
         (_, FnType::FnModule) => {
             unreachable!("methods cannot be FnModule")
@@ -263,11 +280,18 @@ fn ensure_function_options_valid(options: &PyFunctionOptions) -> syn::Result<()>
 }
 
 fn ensure_no_forbidden_protocol_attributes(
+    proto_kind: &PyMethodProtoKind,
     spec: &FnSpec<'_>,
     method_name: &str,
 ) -> syn::Result<()> {
+    if let Some(signature) = &spec.signature.attribute {
+        // __call__ is allowed to have a signature, but nothing else is.
+        if !matches!(proto_kind, PyMethodProtoKind::Call) {
+            bail_spanned!(signature.kw.span() => format!("`signature` cannot be used with magic method `{}`", method_name));
+        }
+    }
     if let Some(text_signature) = &spec.text_signature {
-        bail_spanned!(text_signature.kw.span() => format!("`text_signature` cannot be used with `{}`", method_name));
+        bail_spanned!(text_signature.kw.span() => format!("`text_signature` cannot be used with magic method `{}`", method_name));
     }
     Ok(())
 }
@@ -276,6 +300,7 @@ fn ensure_no_forbidden_protocol_attributes(
 pub fn impl_py_method_def(
     cls: &syn::Type,
     spec: &FnSpec<'_>,
+    doc: &PythonDoc,
     flags: Option<TokenStream>,
 ) -> Result<MethodAndMethodDef> {
     let wrapper_ident = format_ident!("__pymethod_{}__", spec.python_name);
@@ -286,7 +311,7 @@ pub fn impl_py_method_def(
         FnType::FnClass => quote!(Class),
         _ => quote!(Method),
     };
-    let methoddef = spec.get_methoddef(quote! { #cls::#wrapper_ident });
+    let methoddef = spec.get_methoddef(quote! { #cls::#wrapper_ident }, doc);
     let method_def = quote! {
         _pyo3::class::PyMethodDefType::#methoddef_type(#methoddef #add_flags)
     };
@@ -299,10 +324,40 @@ pub fn impl_py_method_def(
 fn impl_py_method_def_new(cls: &syn::Type, spec: &FnSpec<'_>) -> Result<MethodAndSlotDef> {
     let wrapper_ident = syn::Ident::new("__pymethod___new____", Span::call_site());
     let associated_method = spec.get_wrapper_function(&wrapper_ident, Some(cls))?;
+    // Use just the text_signature_call_signature() because the class' Python name
+    // isn't known to `#[pymethods]` - that has to be attached at runtime from the PyClassImpl
+    // trait implementation created by `#[pyclass]`.
+    let text_signature_body = spec.text_signature_call_signature().map_or_else(
+        || quote!(::std::option::Option::None),
+        |text_signature| quote!(::std::option::Option::Some(#text_signature)),
+    );
     let slot_def = quote! {
         _pyo3::ffi::PyType_Slot {
             slot: _pyo3::ffi::Py_tp_new,
-            pfunc: #cls::#wrapper_ident as _pyo3::ffi::newfunc as _
+            pfunc: {
+                unsafe extern "C" fn trampoline(
+                    subtype: *mut _pyo3::ffi::PyTypeObject,
+                    args: *mut _pyo3::ffi::PyObject,
+                    kwargs: *mut _pyo3::ffi::PyObject,
+                ) -> *mut _pyo3::ffi::PyObject
+                {
+                    use _pyo3::impl_::pyclass::*;
+                    impl PyClassNewTextSignature<#cls> for PyClassImplCollector<#cls> {
+                        #[inline]
+                        fn new_text_signature(self) -> ::std::option::Option<&'static str> {
+                            #text_signature_body
+                        }
+                    }
+
+                    _pyo3::impl_::trampoline::newfunc(
+                        subtype,
+                        args,
+                        kwargs,
+                        #cls::#wrapper_ident
+                    )
+                }
+                trampoline
+            } as _pyo3::ffi::newfunc as _
         }
     };
     Ok(MethodAndSlotDef {
@@ -321,7 +376,22 @@ fn impl_call_slot(cls: &syn::Type, mut spec: FnSpec<'_>) -> Result<MethodAndSlot
     let slot_def = quote! {
         _pyo3::ffi::PyType_Slot {
             slot: _pyo3::ffi::Py_tp_call,
-            pfunc: #cls::#wrapper_ident as _pyo3::ffi::ternaryfunc as _
+            pfunc: {
+                unsafe extern "C" fn trampoline(
+                    slf: *mut _pyo3::ffi::PyObject,
+                    args: *mut _pyo3::ffi::PyObject,
+                    kwargs: *mut _pyo3::ffi::PyObject,
+                ) -> *mut _pyo3::ffi::PyObject
+                {
+                    _pyo3::impl_::trampoline::ternaryfunc(
+                        slf,
+                        args,
+                        kwargs,
+                        #cls::#wrapper_ident
+                    )
+                }
+                trampoline
+            } as _pyo3::ffi::ternaryfunc as _
         }
     };
     Ok(MethodAndSlotDef {
@@ -330,29 +400,14 @@ fn impl_call_slot(cls: &syn::Type, mut spec: FnSpec<'_>) -> Result<MethodAndSlot
     })
 }
 
-fn impl_traverse_slot(cls: &syn::Type, spec: FnSpec<'_>) -> MethodAndSlotDef {
-    let ident = spec.name;
+fn impl_traverse_slot(cls: &syn::Type, rust_fn_ident: &syn::Ident) -> MethodAndSlotDef {
     let associated_method = quote! {
         pub unsafe extern "C" fn __pymethod_traverse__(
             slf: *mut _pyo3::ffi::PyObject,
             visit: _pyo3::ffi::visitproc,
             arg: *mut ::std::os::raw::c_void,
-        ) -> ::std::os::raw::c_int
-        {
-            let trap = _pyo3::impl_::panic::PanicTrap::new("uncaught panic inside __traverse__ handler");
-            let pool = _pyo3::GILPool::new();
-            let py = pool.python();
-            let slf = py.from_borrowed_ptr::<_pyo3::PyCell<#cls>>(slf);
-
-            let visit = _pyo3::class::gc::PyVisit::from_raw(visit, arg, py);
-            let borrow = slf.try_borrow();
-            let retval = if let ::std::result::Result::Ok(borrow) = borrow {
-                _pyo3::impl_::pymethods::unwrap_traverse_result(borrow.#ident(visit))
-            } else {
-                0
-            };
-            trap.disarm();
-            retval
+        ) -> ::std::os::raw::c_int {
+            _pyo3::impl_::pymethods::call_traverse_impl::<#cls>(slf, #cls::#rust_fn_ident, visit, arg)
         }
     };
     let slot_def = quote! {
@@ -368,7 +423,7 @@ fn impl_traverse_slot(cls: &syn::Type, spec: FnSpec<'_>) -> MethodAndSlotDef {
 }
 
 fn impl_py_class_attribute(cls: &syn::Type, spec: &FnSpec<'_>) -> syn::Result<MethodAndMethodDef> {
-    let (py_arg, args) = split_off_python_arg(&spec.args);
+    let (py_arg, args) = split_off_python_arg(&spec.signature.arguments);
     ensure_spanned!(
         args.is_empty(),
         args[0].ty.span() => "#[classattr] can only have one argument (of type pyo3::Python)"
@@ -376,9 +431,9 @@ fn impl_py_class_attribute(cls: &syn::Type, spec: &FnSpec<'_>) -> syn::Result<Me
 
     let name = &spec.name;
     let fncall = if py_arg.is_some() {
-        quote!(#cls::#name(py))
+        quote!(function(py))
     } else {
-        quote!(#cls::#name())
+        quote!(function())
     };
 
     let wrapper_ident = format_ident!("__pymethod_{}__", name);
@@ -387,13 +442,10 @@ fn impl_py_class_attribute(cls: &syn::Type, spec: &FnSpec<'_>) -> syn::Result<Me
 
     let associated_method = quote! {
         fn #wrapper_ident(py: _pyo3::Python<'_>) -> _pyo3::PyResult<_pyo3::PyObject> {
+            let function = #cls::#name; // Shadow the method name to avoid #3017
             #deprecations
-            let mut ret = #fncall;
-            if false {
-                use _pyo3::impl_::ghost::IntoPyResult;
-                ret.assert_into_py_result();
-            }
-            _pyo3::callback::convert(py, ret)
+            _pyo3::impl_::pymethods::OkWrap::wrap(#fncall, py)
+                .map_err(::core::convert::Into::into)
         }
     };
 
@@ -413,7 +465,7 @@ fn impl_py_class_attribute(cls: &syn::Type, spec: &FnSpec<'_>) -> syn::Result<Me
 }
 
 fn impl_call_setter(cls: &syn::Type, spec: &FnSpec<'_>) -> syn::Result<TokenStream> {
-    let (py_arg, args) = split_off_python_arg(&spec.args);
+    let (py_arg, args) = split_off_python_arg(&spec.signature.arguments);
 
     if args.is_empty() {
         bail_spanned!(spec.name.span() => "setter function expected to have one argument");
@@ -486,29 +538,34 @@ pub fn impl_py_setter_def(
         }
     };
 
+    let mut cfg_attrs = TokenStream::new();
+    if let PropertyType::Descriptor { field, .. } = &property_type {
+        for attr in field.attrs.iter().filter(|attr| attr.path.is_ident("cfg")) {
+            attr.to_tokens(&mut cfg_attrs);
+        }
+    }
+
     let associated_method = quote! {
-        unsafe extern "C" fn #wrapper_ident(
+        #cfg_attrs
+        unsafe fn #wrapper_ident(
+            _py: _pyo3::Python<'_>,
             _slf: *mut _pyo3::ffi::PyObject,
             _value: *mut _pyo3::ffi::PyObject,
-            _: *mut ::std::os::raw::c_void
-        ) -> ::std::os::raw::c_int {
-            let gil = _pyo3::GILPool::new();
-            let _py = gil.python();
-            _pyo3::callback::panic_result_into_callback_output(_py, ::std::panic::catch_unwind(move || -> _pyo3::PyResult<_> {
-                #slf
-                let _value = _py
-                    .from_borrowed_ptr_or_opt(_value)
-                    .ok_or_else(|| {
-                        _pyo3::exceptions::PyAttributeError::new_err("can't delete attribute")
-                    })?;
-                let _val = _pyo3::FromPyObject::extract(_value)?;
+        ) -> _pyo3::PyResult<::std::os::raw::c_int> {
+            #slf
+            let _value = _py
+                .from_borrowed_ptr_or_opt(_value)
+                .ok_or_else(|| {
+                    _pyo3::exceptions::PyAttributeError::new_err("can't delete attribute")
+                })?;
+            let _val = _pyo3::FromPyObject::extract(_value)?;
 
-                _pyo3::callback::convert(_py, #setter_impl)
-            }))
+            _pyo3::callback::convert(_py, #setter_impl)
         }
     };
 
     let method_def = quote! {
+        #cfg_attrs
         _pyo3::class::PyMethodDefType::Setter({
             #deprecations
             _pyo3::class::PySetterDef::new(
@@ -526,7 +583,7 @@ pub fn impl_py_setter_def(
 }
 
 fn impl_call_getter(cls: &syn::Type, spec: &FnSpec<'_>) -> syn::Result<TokenStream> {
-    let (py_arg, args) = split_off_python_arg(&spec.args);
+    let (py_arg, args) = split_off_python_arg(&spec.signature.arguments);
     ensure_spanned!(
         args.is_empty(),
         args[0].ty.span() => "getter function can only have one argument (of type pyo3::Python)"
@@ -558,7 +615,6 @@ pub fn impl_py_getter_def(
             ..
         } => {
             // named struct field
-            //quote!(_slf.#ident.clone())
             quote!(::std::clone::Clone::clone(&(_slf.#ident)))
         }
         PropertyType::Descriptor { field_index, .. } => {
@@ -610,22 +666,27 @@ pub fn impl_py_getter_def(
         }
     };
 
+    let mut cfg_attrs = TokenStream::new();
+    if let PropertyType::Descriptor { field, .. } = &property_type {
+        for attr in field.attrs.iter().filter(|attr| attr.path.is_ident("cfg")) {
+            attr.to_tokens(&mut cfg_attrs);
+        }
+    }
+
     let associated_method = quote! {
-        unsafe extern "C" fn #wrapper_ident(
-            _slf: *mut _pyo3::ffi::PyObject,
-            _: *mut ::std::os::raw::c_void
-        ) -> *mut _pyo3::ffi::PyObject {
-            let gil = _pyo3::GILPool::new();
-            let _py = gil.python();
-            _pyo3::callback::panic_result_into_callback_output(_py, ::std::panic::catch_unwind(move || -> _pyo3::PyResult<_> {
-                #slf
-                let item = #getter_impl;
-                #conversion
-            }))
+        #cfg_attrs
+        unsafe fn #wrapper_ident(
+            _py: _pyo3::Python<'_>,
+            _slf: *mut _pyo3::ffi::PyObject
+        ) -> _pyo3::PyResult<*mut _pyo3::ffi::PyObject> {
+            #slf
+            let item = #getter_impl;
+            #conversion
         }
     };
 
     let method_def = quote! {
+        #cfg_attrs
         _pyo3::class::PyMethodDefType::Getter({
             #deprecations
             _pyo3::class::PyGetterDef::new(
@@ -659,6 +720,7 @@ pub enum PropertyType<'a> {
     Function {
         self_type: &'a SelfType,
         spec: &'a FnSpec<'a>,
+        doc: PythonDoc,
     },
 }
 
@@ -693,7 +755,7 @@ impl PropertyType<'_> {
             PropertyType::Descriptor { field, .. } => {
                 Cow::Owned(utils::get_doc(&field.attrs, None))
             }
-            PropertyType::Function { spec, .. } => Cow::Borrowed(&spec.doc),
+            PropertyType::Function { doc, .. } => Cow::Borrowed(doc),
         }
     }
 }
@@ -1039,7 +1101,6 @@ impl SlotDef {
         let py = syn::Ident::new("_py", Span::call_site());
         let arg_types: &Vec<_> = &arguments.iter().map(|arg| arg.ffi_type()).collect();
         let arg_idents: &Vec<_> = &(0..arguments.len())
-            .into_iter()
             .map(|i| format_ident!("arg{}", i))
             .collect();
         let wrapper_ident = format_ident!("__pymethod_{}__", method_name);
@@ -1052,22 +1113,36 @@ impl SlotDef {
             *extract_error_mode,
             return_mode.as_ref(),
         )?;
+        let name = spec.name;
         let associated_method = quote! {
-            unsafe extern "C" fn #wrapper_ident(_raw_slf: *mut _pyo3::ffi::PyObject, #(#arg_idents: #arg_types),*) -> #ret_ty {
+            unsafe fn #wrapper_ident(
+                #py: _pyo3::Python<'_>,
+                _raw_slf: *mut _pyo3::ffi::PyObject,
+                #(#arg_idents: #arg_types),*
+            ) -> _pyo3::PyResult<#ret_ty> {
+                let function = #cls::#name; // Shadow the method name to avoid #3017
                 let _slf = _raw_slf;
-                let gil = _pyo3::GILPool::new();
-                let #py = gil.python();
-                _pyo3::callback::panic_result_into_callback_output(#py, ::std::panic::catch_unwind(move || -> _pyo3::PyResult<_> {
-                    #body
-                }))
+                #body
             }
         };
-        let slot_def = quote! {
+        let slot_def = quote! {{
+            unsafe extern "C" fn trampoline(
+                _slf: *mut _pyo3::ffi::PyObject,
+                #(#arg_idents: #arg_types),*
+            ) -> #ret_ty
+            {
+                _pyo3::impl_::trampoline:: #func_ty (
+                    _slf,
+                    #(#arg_idents,)*
+                    #cls::#wrapper_ident
+                )
+            }
+
             _pyo3::ffi::PyType_Slot {
                 slot: _pyo3::ffi::#slot,
-                pfunc: #cls::#wrapper_ident as _pyo3::ffi::#func_ty as _
+                pfunc: trampoline as _pyo3::ffi::#func_ty as _
             }
-        };
+        }};
         Ok(MethodAndSlotDef {
             associated_method,
             slot_def,
@@ -1084,9 +1159,10 @@ fn generate_method_body(
     return_mode: Option<&ReturnMode>,
 ) -> Result<TokenStream> {
     let self_conversion = spec.tp.self_conversion(Some(cls), extract_error_mode);
+    let self_arg = spec.tp.self_arg();
     let rust_name = spec.name;
     let args = extract_proto_arguments(py, spec, arguments, extract_error_mode)?;
-    let call = quote! { _pyo3::callback::convert(#py, #cls::#rust_name(_slf, #(#args),*)) };
+    let call = quote! { _pyo3::callback::convert(#py, #cls::#rust_name(#self_arg #(#args),*)) };
     let body = if let Some(return_mode) = return_mode {
         return_mode.return_call_output(py, call)
     } else {
@@ -1138,7 +1214,6 @@ impl SlotFragmentDef {
         let py = syn::Ident::new("_py", Span::call_site());
         let arg_types: &Vec<_> = &arguments.iter().map(|arg| arg.ffi_type()).collect();
         let arg_idents: &Vec<_> = &(0..arguments.len())
-            .into_iter()
             .map(|i| format_ident!("arg{}", i))
             .collect();
         let body = generate_method_body(cls, spec, &py, arguments, *extract_error_mode, None)?;
@@ -1231,10 +1306,10 @@ fn extract_proto_arguments(
     proto_args: &[Ty],
     extract_error_mode: ExtractErrorMode,
 ) -> Result<Vec<TokenStream>> {
-    let mut args = Vec::with_capacity(spec.args.len());
+    let mut args = Vec::with_capacity(spec.signature.arguments.len());
     let mut non_python_args = 0;
 
-    for arg in &spec.args {
+    for arg in &spec.signature.arguments {
         if arg.py {
             args.push(quote! { #py });
         } else {
